
import sys
import logging
from threading import Thread
from argparse import Namespace
from functools import partial


class Logger:
    """
        Handle logging with Python's built-in logging facilities simplified
    """
    def __init__(self, log_filename=None, logfile_mode='a', logfile_encoding='UTF-8', logfile_level='INFO',
                 console_stream=sys.stderr, console_level='INFO', console_format='{asctime} {levelname}: {message}',
                 file_format='{asctime} {levelname}: {message}'):
        # logging.basicConfig(level=logging.INFO)  # For debugging requests
        log_levels = {'DEBUG': logging.DEBUG, 'INFO': logging.INFO, 'WARNING': logging.WARNING, 'ERROR': logging.ERROR,
                      'CRITICAL': logging.CRITICAL}

        if console_level not in log_levels:
            raise KeyError('Console loglevel is not valid ({0}): {1}'.format(', '.join(log_levels.keys()),
                                                                             console_level))
        console_level = log_levels[console_level]

        if logfile_level not in log_levels:
            raise KeyError('Logfile loglevel is not valid ({0}): {1}'.format(', '.join(log_levels.keys()),
                                                                             logfile_level))
        logfile_level = log_levels[logfile_level]

        # Create logger
        self._logger = logging.getLogger(log_filename)  # Logger is named after the logfile
        self._logger.propagate = False

        # Create one handler for console output and set its properties accordingly
        c_handler = logging.StreamHandler(stream=console_stream)
        c_handler.setLevel(console_level)

        # Create formatters and add them to handlers
        c_format = logging.Formatter(console_format, style='{')
        c_handler.setFormatter(c_format)

        # Add handlers to the logger
        self._logger.addHandler(c_handler)

        # Create another handler for the logfile and set its properties accordingly
        if log_filename is not None:
            f_handler = logging.FileHandler(log_filename, mode=logfile_mode, encoding=logfile_encoding)
            f_handler.setLevel(logfile_level)
            f_format = logging.Formatter(file_format, style='{')
            f_handler.setFormatter(f_format)
            self._logger.addHandler(f_handler)

        self._logger.setLevel(min(console_level, logfile_level))

        self._leveled_logger = {'DEBUG': self._logger.debug, 'INFO': self._logger.info, 'WARNING': self._logger.warning,
                                'ERROR': self._logger.error, 'CRITICAL': self._logger.critical}

        # Variables for supporting multiprocess logging
        self._q = None
        self._lp = None

        self.log('INFO', 'Logging started')

    def log(self, level, *message, sep=' ', end='\n', file=None):
        """
            A print()-like logging function
                :param level: (str) Levels from the standard set: 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'
                :param message: One or more elems as for print()
                :param sep: Separator element as for print()
                :param end: Ending element as for print()
                :param file: Ignored as handlers are set in __init__()
                :return: None
        """
        _ = file  # Silence IDE
        for handler in self._logger.handlers:
            handler.terminator = end
        if level not in self._leveled_logger:
            self._leveled_logger['CRITICAL']('UNKNOWN LOGGING LEVEL SPECIFIED FOR THE NEXT ENTRY: {0}'.format(level))
            level = 'CRITICAL'
        self._leveled_logger[level](sep.join(str(msg) for msg in message))

    @staticmethod
    def _log_to_queue(queue, level, *args, **kwargs):
        queue.put((level, args, kwargs))

    @staticmethod
    def _logger_thread(queue, log_fun):
        while True:
            record = queue.get()
            if record is None:
                break
            level, log_args, log_kwargs = record
            log_fun(level, *log_args, **log_kwargs)  # LOG HERE from init parameter!

    def init_mp_logging_context(self, queue):
        """Create a separate thread for logging through a Queue (see the following example)

            from itertools import repeat
            from multiprocessing import Pool, Manager, current_process

            from logger import Logger

            def worker_process(par):
                state, lq = par
                retl = []
                for n in range(100):
                    lq.log('WARNING', f'{current_process().name} message{state} {n}')
                lq.log('WARNING', f'{current_process().name} finished')
                return retl

            log_obj = Logger('test.log', 'w')  # Apply all parameters for logger here!
            log_obj.log('INFO', 'NORMAL LOG BEGIN')  # Normal logging
            with Manager() as man:
                log_queue = man.Queue()
                with log_obj.init_mp_logging_context(log_queue) as mplogger, Pool() as p:
                    # Here one can log parallel from all processes!
                    return_queue = p.imap(worker_process, zip(range(10), repeat(mplogger)), chunksize=3)
                    for _ in return_queue:
                        pass
            log_obj.log('INFO', 'NORMAL LOG END')  # Normal logging
        """
        self._q = queue
        self._lp = Thread(target=self._logger_thread, args=(self._q, self.log))
        self._lp.start()
        return self

    def __enter__(self):
        """Enter into context with prepared queue"""
        if self._q is None:
            raise RuntimeError('Must call init_mp_logging_context() with an initialised Queue as param'
                               ' before entering into context!')
        return Namespace(log=partial(self._log_to_queue, self._q))

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Tell the logging thread to finish up"""
        self._q.put(None)
        self._lp.join()

    def __del__(self):
        handlers = list(self._logger.handlers)  # Copy, because we write the same variable in the loop!
        for h in handlers:
            self._logger.removeHandler(h)
            h.flush()
            if isinstance(h, logging.FileHandler):
                h.close()



#!/usr/bin/env python3
# -*- coding: utf-8, vim: expandtab:ts=4 -*-

import sys
from argparse import ArgumentParser, ArgumentTypeError, FileType

from webarticlecurator import wrap_input_constants, NewsArchiveCrawler, NewsArticleCrawler, sample_warc_by_urls, \
    validate_warc_file, online_test, Logger, __version__


def str2bool(v):
    """
    Original code from:
     https://stackoverflow.com/questions/15008758/parsing-boolean-values-with-argparse/43357954#43357954
    """
    if v.lower() in ('yes', 'Yes', 'YES', 'true', 'True', 'TRUE', 't', 'T', 'y', 'Y', '1'):
        return True
    elif v.lower() in ('no', 'No', 'NO', 'false', 'False', 'FALSE', 'f', 'F', 'n', 'N', '0'):
        return False
    else:
        raise ArgumentTypeError('Boolean value expected.')


def parse_args_crawl(parser):
    parser.add_argument(dest='command', choices={'crawl'}, metavar='crawl',
                        help='Crawl a portal with the supplied configuation and arguments')
    parser.add_argument('config', type=str, help='Portal configfile (see configs folder for examples!)')
    parser.add_argument('--old-archive-warc', type=str, help='Existing WARC archives of the portal\'s archive '
                                                             '(Use them as cache)', nargs='+', default=None)
    parser.add_argument('--archive-warc', type=str, help='New WARC archive of the portal\'s archive '
                                                         '(Copy all cached pages if --old-archive-warc is specified)')
    parser.add_argument('--old-articles-warc', type=str, help='Existing WARC archives of the portal\'s archive '
                                                              '(Use them as cache)', nargs='+', default=None)
    parser.add_argument('--articles-warc', type=str, help='New WARC archive of the portal\'s archive '
                                                          '(Copy all cached pages if --old-archive-warc is specified)')
    parser.add_argument('--archive-just-cache', type=str2bool, nargs='?', const=True, default=False,
                        metavar='True/False', help='Use only cached pages (no output warcfile):'
                                                   ' --old-archive-warc must be specified!')
    parser.add_argument('--articles-just-cache', type=str2bool, nargs='?', const=True, default=False,
                        metavar='True/False', help='Use only cached pages (no output warcfile):'
                                                   ' --old-articles-warc must be specified!')
    parser.add_argument('--debug-news-archive', type=str2bool, nargs='?', const=True, default=False,
                        metavar='True/False', help='Set DEBUG logging on NewsArchiveCrawler'
                                                   ' and print the number of extracted URLs per page')
    parser.add_argument('--strict', type=str2bool, nargs='?', const=True, default=False, metavar='True/False',
                        help='Set strict-mode in WARCReader to enable validation')
    parser.add_argument('--crawler-name', type=str, help='The name of the crawler for the WARC info record',
                        default='WebArticleCurator {0}'.format(__version__))
    parser.add_argument('--user-agent', type=str, help='The User-Agent string to use in headers while downloading')
    parser.add_argument('--no-overwrite-warc', help='Do not overwrite --{archive,articles}-warc if needed',
                        action='store_false')
    parser.add_argument('--cumulative-error-threshold', type=int, help='Sum of download errors before giving up',
                        default=15)
    parser.add_argument('--known-bad-urls', type=str, help='Known bad URLs to be excluded from download (filename, '
                                                           'one URL per line)', default=None)
    parser.add_argument('--known-article-urls', type=str, help='Known article URLs to mark the desired end of '
                                                               'the archive (filename, one URL per line)', default=None)
    parser.add_argument('--max-no-of-calls-in-period', type=int, help='Limit number of HTTP request per period',
                        default=2)
    parser.add_argument('--limit-period', type=int, help='Limit (seconds) the period the number of HTTP request'
                                                         ' see also --max-no-of-calls-in-period',
                        default=1)
    parser.add_argument('--proxy-url', type=str, help='SOCKS Proxy URL to use eg. socks5h://localhost:9050',
                        default=None)
    parser.add_argument('--allow-cookies', type=str2bool, nargs='?', const=True, default=False, metavar='True/False',
                        help='Allow session cookies')
    parser.add_argument('--stay-offline', type=str2bool, nargs='?', const=True, default=False, metavar='True/False',
                        help='Do not download, but write output WARC (see --just-cache when no output warcfile'
                             ' is needed)')

    # Mutually exclusive group...
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--archive', help='Crawl only the portal\'s archive', action='store_true')
    group.add_argument('--articles', help='Crawl articles (and optionally use cached WARC for the portal\'s archive),'
                                          ' DEFAULT behaviour', action='store_true')
    group.add_argument('--corpus', help='Use --old-articles-warc to create corpus (no crawling, equals to'
                                        ' --archive-just-cache and --articles-just-cache)', action='store_true')
    cli_args = parser.parse_args()
    # If archive is True -> articles is False, if archive is False -> articles is True
    cli_args.articles = not cli_args.archive
    if cli_args.corpus:
        cli_args.archive_just_cache = True
        cli_args.articles_just_cache = True
    if cli_args.archive and (not cli_args.archive_warc and not cli_args.archive_just_cache):
        print('Must specify at least --archive-warc as destination!', file=sys.stderr)
        exit(1)
    if cli_args.articles and ((not cli_args.archive_warc and not cli_args.archive_just_cache) or
                              (not cli_args.articles_warc and not cli_args.articles_just_cache)):
        print('Must specify at least --archive-warc and --articles-warc as destination!', file=sys.stderr)
        exit(1)
    if cli_args.corpus and not cli_args.old_articles_warc:
        print('Must specify at least --old-articles-warc as source!', file=sys.stderr)
        exit(1)

    if cli_args.debug_news_archive:
        cli_args.debug_params = {'console_level': 'DEBUG', 'logfile_level': 'DEBUG'}
    else:
        cli_args.debug_params = {}
    return cli_args


def parse_args_validate_and_list(parser):
    parser.add_argument('command', choices={'validate', 'listurls'}, metavar='validate|listurls',
                        help='Validate a warc file (created by this program) or list the urls in it')
    parser.add_argument('-s', '--source-warcfile', type=str, metavar='SOURCE WARCFILE', nargs='+',
                        help='A warc file (created by this program) to work from')
    args = parser.parse_args()
    if (args.source_warcfile is None or len(args.source_warcfile) == 0) and args.offline:
        print('Must specify at least one SOURCE_WARC !', file=sys.stderr)
        exit(1)
    return args


def parse_args_sample(parser):
    parser.add_argument(dest='command', choices={'sample'}, metavar='sample',
                        help='Copy the supplied list of URLs to the output warc file from the internet '
                             'or from the supplied warc file (created by this program)')
    parser.add_argument('-s', '--source-warcfile', type=str, default=None, nargs='*', metavar='SOURCE WARCFILE',
                        help='A warc file (created by this program) to work from '
                             '(not mandatory when --offline is True)')
    parser.add_argument('-i', '--input-urls', dest='url_input_stream', type=FileType(), default=sys.stdin,
                        help='Use input file instead of STDIN (one URL per line)', metavar='FILE', required=False)
    parser.add_argument('target_warcfile', type=str, metavar='TARGET_WARFCILE', help='The name of the target warc file')
    parser.add_argument('--offline', type=str2bool, nargs='?', const=True, default=True, metavar='True/False',
                        help='Download URLs which are not present in the source archive (default True)')
    args = parser.parse_args()
    if (args.source_warcfile is None or len(args.source_warcfile) == 0) and args.offline:
        print('Must specify at least one SOURCE_WARC if --offline is False!', file=sys.stderr)
        exit(1)
    return args


def parse_args_cat(parser):
    parser.add_argument(dest='command', choices={'cat'}, metavar='cat',
                        help='Print the list of URLs from the supplied warc file (created by this program)')
    parser.add_argument('-s', '--source-warcfile', type=str, metavar='SOURCE WARCFILE', nargs='+',
                        help='A warc file (created by this program) to work from')
    parser.add_argument('-i', '--input-urls', dest='url_input_stream', type=FileType(), default=sys.stdin,
                        help='Use input file instead of STDIN (one URL per line)', metavar='FILE')
    parser.add_argument('out_dir', type=str)
    args = parser.parse_args()
    if (args.source_warcfile is None or len(args.source_warcfile) == 0) and args.offline:
        print('Must specify at least one SOURCE_WARC !', file=sys.stderr)
        exit(1)
    return args


def parse_args_donwload(parser):
    parser.add_argument(dest='command', choices={'download'}, metavar='downlaod',
                        help='Download a single URL to a warc file')
    parser.add_argument('source_url', type=str, metavar='URL', help='The URL to download')
    parser.add_argument('target_warcfile', type=str, metavar='TARGET_WARFCILE', help='The name of the target warc file')
    return parser.parse_args()


def main_crawl(args):
    """ read input data from the given files, initialize variables """
    portal_settings = wrap_input_constants(args.config)
    # These parameters go down directly to the downloader
    download_params = {'program_name': args.crawler_name, 'user_agent': args.user_agent,
                       'overwrite_warc': args.no_overwrite_warc, 'err_threshold': args.cumulative_error_threshold,
                       'known_bad_urls': args.known_bad_urls, 'strict_mode': args.strict,
                       'max_no_of_calls_in_period': args.max_no_of_calls_in_period, 'limit_period': args.limit_period,
                       'proxy_url': args.proxy_url, 'allow_cookies': args.allow_cookies,
                       'stay_offline': args.stay_offline, 'verify_request': portal_settings['verify_request']}
    if args.archive:
        # For the article links only...
        archive_crawler = NewsArchiveCrawler(portal_settings, args.old_archive_warc, args.archive_warc,
                                             args.archive_just_cache, args.known_article_urls, args.debug_params,
                                             download_params)
        for url in archive_crawler.url_iterator():  # Get the list of urls in the archive...
            print(url, flush=True)
    else:
        articles_crawler = NewsArticleCrawler(portal_settings, args.old_articles_warc, args.articles_warc,
                                              args.old_archive_warc, args.archive_warc, args.articles_just_cache,
                                              args.archive_just_cache, args.known_article_urls, args.debug_params,
                                              download_params)
        articles_crawler.download_and_extract_all_articles()


def main_validate_and_list(args):
    """ __file__ validate [source warcfiles]     # WarcReader(..., strict_mode=True, check_digest=True) """
    level = 'INFO'
    url_index = validate_warc_file(args.source_warcfile, Logger(console_level=level, logfile_level=level))
    if args.command == 'listurls':
        for url in url_index:
            print(url)


def main_cat_and_sample(args):
    """ __file__ sample [source warcfiles or None] [urls list file or stdin] [target warcfile] [Online or Offline] """
    main_logger = Logger()
    out_dir = getattr(args, 'out_dir', None)
    target_warcfile = getattr(args, 'target_warcfile', None)
    target = out_dir if out_dir is not None else target_warcfile
    main_logger.log('INFO', 'Adding URLs to', target, ':')
    offline = getattr(args, 'offline', True)  # Sample can be online or offline, but we write warc only when sampling!
    sample_warc_by_urls(args.source_warcfile, args.url_input_stream, main_logger, target_warcfile=target_warcfile,
                        offline=offline, out_dir=out_dir, just_cache=args.command == 'cat')
    main_logger.log('INFO', 'Done!')


def main_download(args):
    """ __file__ download [URL] [target warfile] """
    main_logger = Logger()
    main_logger.log('INFO', 'Adding URL to', args.target_warcfile, ':')
    online_test(args.source_url, args.target_warcfile, main_logger)
    main_logger.log('INFO', 'Done!')


def main():
    # Parse command arg from CLI and pass to the selected main function
    commands = {'validate': (parse_args_validate_and_list, main_validate_and_list),
                'listurls': (parse_args_validate_and_list, main_validate_and_list),
                'sample': (parse_args_sample, main_cat_and_sample), 'download': (parse_args_donwload, main_download),
                'cat': (parse_args_cat, main_cat_and_sample), 'crawl': (parse_args_crawl, main_crawl)}
    parser = ArgumentParser()
    parser.add_argument('command', choices=commands.keys(), metavar='COMMAND',
                        help='Please choose from the available commands ({0}) to set mode and see deatiled help!'.
                        format(set(commands.keys())))

    command = parser.parse_args(sys.argv[1:2]).command  # Route ArgumentParser before reparsing the whole CLI
    argparse_fun, main_fun = commands[command]
    sub_args = argparse_fun(ArgumentParser())
    main_fun(sub_args)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8, vim: expandtab:ts=4 -*-

import os
import sys
import importlib.util
from argparse import Namespace
from datetime import datetime, date, timedelta

import yamale

site_schema = yamale.make_schema(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'site_schema.yaml'))
crawl_schema = yamale.make_schema(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'crawl_schema.yaml'))


def load_and_validate(schema, fname):
    data = yamale.make_data(fname)
    try:
        yamale.validate(schema, data, strict=True)
    except yamale.YamaleError as e:
        for result in e.results:
            print('Error validating data {0} with {1}:'.format(result.data, result.schema), file=sys.stderr)
            for error in result.errors:
                print('', error, sep='\t', file=sys.stderr)
        exit(1)
    return data[0][0]


def import_python_file(module_name, file_path):
    # Import module from file: https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


def wrap_input_constants(current_task_config_filename):
    """
        Helper to store and process input data so that main function does not contain so many
         codelines of variable initialization
        Fields should be handled as constants after initialization
         CAPITALIZED KEYS are transformed runtime (e.g. Regular Expressions),
         lowercase keys are present in the config and will be used as is
    """
    # Instructions to the current task
    settings = load_and_validate(crawl_schema, current_task_config_filename)

    # The directory name of the configs
    settings['DIR_NAME'] = os.path.dirname(os.path.abspath(current_task_config_filename))

    # Technical data about the website to crawl
    site_schema_fname = os.path.join(settings['DIR_NAME'], settings['schema'])
    settings['SITE_SCHEMA_DIR_NAME'] = os.path.dirname(os.path.abspath(site_schema_fname))
    current_site_schema = load_and_validate(site_schema, site_schema_fname)
    settings.update(current_site_schema)

    # Date filtering ON in any other cases OFF
    settings['FILTER_ARTICLES_BY_DATE'] = 'date_from' in settings and 'date_until' in settings

    for column in settings['columns'].keys():
        column_settings = settings['columns'][column]
        # We use the supplied dates (None != date_first_article <= date_from <= date_until < today (or raise exception))
        # to generate all URLs from the first day to the last one-by-one
        if 'date_from' in settings:
            column_settings['DATE_FROM'] = settings['date_from']
        elif 'date_first_article' in column_settings:
            column_settings['DATE_FROM'] = column_settings['date_first_article']

        if 'date_until' in settings:
            column_settings['DATE_UNTIL'] = settings['date_until']
        elif'date_last_article' in column_settings:
            column_settings['DATE_UNTIL'] = column_settings['date_last_article']
        else:
            column_settings['DATE_UNTIL'] = date.today() - timedelta(1)  # yesterday

        if settings['FILTER_ARTICLES_BY_DATE'] or settings['archive_page_urls_by_date']:
            if 'DATE_FROM' not in column_settings:
                raise ValueError('DateError: date_first_article and date_from is not set please'
                                 ' set at least one of them!')
            if not (column_settings['DATE_FROM'] <= column_settings['DATE_UNTIL'] < date.today()):
                raise ValueError('DateError: DATE_FROM ({0}) <= DATE UNTIL ({1}) < date.doday() ({2}) is not'
                                 ' satisfiable! Please check date_from ({3}), date_until ({4}),'
                                 ' date_first_article ({5}) and date_last_article ({6})!'.
                                 format(column_settings['DATE_FROM'], column_settings['DATE_UNTIL'], date.today(),
                                        settings['date_from'], settings['date_until'],
                                        column_settings['date_first_article'], column_settings['date_last_article']))

        min_pagenum = column_settings.get('min_pagenum', -1)  # Can not normally be -1!
        initial_pagenum = column_settings.get('initial_pagenum', '')
        max_pagenum = column_settings.get('max_pagenum')

        if settings['next_url_by_pagenum']:
            # If column consists of only one page, min_pagenum and max_pagenum must be omited
            # while initial_pagenum must be set
            if 'min_pagenum' not in column_settings:
                if 'max_pagenum' in column_settings or 'initial_pagenum' not in column_settings:
                    raise ValueError('min_pagenum can be omited iff max_pagenum is not present'
                                     ' and initial_pagenum is set, when next_url_by_pagenum is true!')
                else:  # min_pagenum and max_pagenum are not set, but initial_pagenum is. -> One page column!
                    min_pagenum = 2  # so max < min and exit immediately
                    max_pagenum = 1
            else:
                # If initial_pagenum is explicitly set as int and min_pagenum is set,
                #  initial_pagenum + 1 == min_pagenum <= max_pagenum must be satisfied!
                if ((isinstance(initial_pagenum, int) and min_pagenum != initial_pagenum + 1) or
                   (isinstance(max_pagenum, int) and min_pagenum > max_pagenum)):
                    raise ValueError('If two or more from initial_pagenum, min_pagenum and max_pagenum are set,'
                                     ' initial_pagenum + 1 == min_pagenum <= max_pagenum must be satisfied!')

        # If initial_pagenum is implicit, then it will be substituted with empty string. e.g. in &page=
        column_settings['INITIAL_PAGENUM'] = str(initial_pagenum)
        column_settings['min_pagenum'] = min_pagenum
        column_settings['max_pagenum'] = max_pagenum

    settings['new_article_url_threshold'] = settings.get('new_article_url_threshold')

    # Portal specific functions
    file_path = settings['portal_specific_exctractor_functions_file']
    module = import_python_file('portal_specific_exctractor_functions',
                                os.path.join(settings['SITE_SCHEMA_DIR_NAME'], file_path))
    for attr_name, attr_name_dest, mandatory in \
            (('extract_next_page_url_fun', 'EXTRACT_NEXT_PAGE_URL_FUN', False),
             ('extract_article_urls_from_page_fun', 'EXTRACT_ARTICLE_URLS_FROM_PAGE_FUN', True),
             ('next_page_of_article_fun', 'NEXT_PAGE_OF_ARTICLE_FUN', False)):
        settings[attr_name_dest] = getattr(module, settings.get(attr_name, ''), None)
        if mandatory and settings[attr_name_dest] is None:
            raise ValueError('{0} is unset!'.format(attr_name))

    if settings['infinite_scrolling']:
        if not settings['next_url_by_pagenum']:
            raise ValueError('If infinite_scrolling is true, next_url_by_pagenum must be also true!')
        if settings['EXTRACT_NEXT_PAGE_URL_FUN'] is not None:
            raise ValueError('If infinite_scrolling is true, extract_next_page_url_fun must be None!')

    # Set and init converter class which is dummy-converter by default
    corp_conv = settings.get('corpus_converter', 'dummy-converter')
    if corp_conv == 'dummy-converter':
        if settings['NEXT_PAGE_OF_ARTICLE_FUN'] is not None:
            raise ValueError('If next_page_of_article_fun is supplied corpus_converter cannot be dummy-converter!')
        corpus_converter_class = DummyConverter
        if settings['FILTER_ARTICLES_BY_DATE'] and not settings['archive_page_urls_by_date']:
            raise ValueError('Date filtering is not possible with DummyConverter with a non-date-based archive!')
        settings['FILTER_ARTICLES_BY_DATE'] = False  # Use the archive dates for filtering...
    else:
        file_path = settings['corpus_converter_file']
        if file_path is None:
            raise ValueError('corpus_converter is {0}, but {1} is unset!'.format(corp_conv, file_path))
        module = import_python_file('corpus_converter', os.path.join(settings['SITE_SCHEMA_DIR_NAME'], file_path))
        corpus_converter_class = getattr(module, corp_conv)
    settings['CORPUS_CONVERTER'] = corpus_converter_class(settings)

    return settings


class DummyConverter:  # No output corpus
    """
        An example converter to showcase API and to suppress any article processing at crawling time (for new portals)
    """

    def __init__(self, settings):
        self._logger = Namespace(log=print)  # Hack to be able to monkeypatch logger
        # Override this if needed!
        if settings['FILTER_ARTICLES_BY_DATE'] and not settings['archive_page_urls_by_date']:
            raise ValueError(f'Date filtering is not possible with {type(self).__name__} on a non-date-based archive!')
        settings['FILTER_ARTICLES_BY_DATE'] = False  # Use the archive dates for filtering...
        # Init stuff
        _ = settings  # Silence IDE

    @staticmethod
    def identify_site_scheme(url, article_raw_html):
        _ = url, article_raw_html  # Silence IDE

    @staticmethod
    def extract_article_date(url, article_raw_html, scheme):
        """ extracts and returns next page URL from an HTML code if there is one... """
        _ = url, article_raw_html, scheme  # Silence dummy IDE
        return datetime.today().date()

    @staticmethod
    def article_to_corpus(url, article_raw_html, scheme):
        _ = url, article_raw_html, scheme  # Silence dummy IDE
        pass

    @staticmethod
    def follow_links_on_page(url, article_raw_html, scheme):
        _ = url, article_raw_html, scheme  # Silence dummy IDE
        return set()

    def __del__(self):
        pass


if __name__ == '__main__':
    if len(sys.argv) == 2:
        wrap_input_constants(sys.argv[1])
        print(f'Config {sys.argv[1]} OK!')



#!/usr/bin/env python3
# -*- coding: utf-8, vim: expandtab:ts=4 -*-

from datetime import timedelta
from calendar import monthrange, isleap

from webarticlecurator import WarcCachingDownloader, Logger


def add_and_write_factory(attr, fname):
    """A helper function to write gathered URLs to a file handle if it is supplied"""
    if fname is None:
        return None, attr.add
    else:
        fh = open(fname, 'w', encoding='UTF-8')  # To store FH (for closing it)

        def add_fun(elem):
            attr.add(elem)
            print(elem, file=fh, flush=True)

        return fh, add_fun


class NewsArchiveCrawler:
    """
        Using the provided regexes
        1) Generates URLs of lists of articles (archives)
        2) Extracts URLs of articles from these lists (with helper functions and config)
    """
    def __init__(self, settings, existing_archive_filenames, new_archive_filename, archive_just_cache=False,
                 known_article_urls=None, debug_params=None, downloader_params=None):

        # List the used properties
        self._archive_page_urls_by_date = None
        self._archive_url_format = None
        self._date_from = None
        self._date_until = None
        self._go_reverse_in_archive = None
        self._min_pagenum = None
        self._initial_page_num = None
        self._ignore_archive_cache = None
        self._infinite_scrolling = None
        self._extract_article_urls_from_page_fun = None
        self._find_next_page_url = None

        # Save the original settings for using it with all columns
        self._settings = settings
        # Get columns, if there is only one we use None to keep default settings
        self._columns = settings['columns']

        # Initialise the logger
        if debug_params is None:
            debug_params = {}
        self._logger = Logger(settings['log_file_archive'], **debug_params)

        # Open files for writing gathered URLs if needed,
        self.good_urls = set()
        self._new_good_archive_urls_fh, self._good_urls_add = \
            add_and_write_factory(self.good_urls, settings.get('new_good_archive_urls'))

        self.problematic_urls = set()
        self._new_problematic_archive_urls_fh, self._problematic_urls_add = \
            add_and_write_factory(self.problematic_urls, settings.get('new_problematic_archive_urls'))

        # Setup the list of cached article URLs to stop archive crawling in time
        self.known_article_urls = set()
        if known_article_urls is not None:
            if isinstance(known_article_urls, str):
                with open(known_article_urls, encoding='UTF-8') as fh:
                    self.known_article_urls = {line.strip() for line in fh}
            elif isinstance(known_article_urls, set):
                self.known_article_urls = known_article_urls

        # Create new archive while downloading, or simulate download and read the archive
        self._downloader = WarcCachingDownloader(existing_archive_filenames, new_archive_filename, self._logger,
                                                 archive_just_cache, downloader_params)
        # Known bad URLs (read-only, available at __init__ time)
        self.bad_urls = self._downloader.bad_urls
        # Known good URLs (read-only, available at __init__ time from cache)
        self.url_index = self._downloader.url_index

    def _store_settings(self, column_spec_settings):
        # Settings for URL iterator
        self._archive_page_urls_by_date = self._settings['archive_page_urls_by_date']
        self._archive_url_format = column_spec_settings['archive_url_format']
        if self._archive_page_urls_by_date:
            self._date_from = column_spec_settings['DATE_FROM']
            self._date_until = column_spec_settings['DATE_UNTIL']
            self._go_reverse_in_archive = self._settings['go_reverse_in_archive']

        # Settings for gen_article_urls_including_subpages()
        self._min_pagenum = column_spec_settings['min_pagenum']
        self._initial_page_num = column_spec_settings['INITIAL_PAGENUM']
        self._ignore_archive_cache = self._settings['ignore_archive_cache']
        self._infinite_scrolling = self._settings['infinite_scrolling']
        self._extract_article_urls_from_page_fun = self._settings['EXTRACT_ARTICLE_URLS_FROM_PAGE_FUN']

        # Store the constant parameters for the actual function used later
        self._find_next_page_url = \
            self._find_next_page_url_factory(self._settings['EXTRACT_NEXT_PAGE_URL_FUN'],
                                             self._settings['next_url_by_pagenum'],
                                             self._settings['infinite_scrolling'], column_spec_settings['max_pagenum'],
                                             self._settings['new_article_url_threshold'], self.known_article_urls)

    def __del__(self):  # Write newly found URLs to files when output files supplied...
        # Save the good URLs...

        if hasattr(self, '_new_good_archive_urls_fh') and hasattr(self, 'close'):
            self._new_good_archive_urls_fh.close()

        if hasattr(self, '_new_problematic_archive_urls_fh') and hasattr(self, 'close'):
            self._new_problematic_archive_urls_fh.close()

    def url_iterator(self):
        """
            The URL generation logic. We have one or more base URL to the archive (or column archives if there is more)
                and it is complete with portal-specific ending.
            The archive can be stored in groups (mostly the creation date) or ordered in a flat list paginated.
            This two main method can also be mixed.
            Pagination can be implemented in various vays see the appropriate function for details
        :return: Every page of the archive contain multiple URL to the actual articles, which are extrated and
         then returned as an iterator based on URLs.
        """
        for column_name, params in self._columns.items():
            self._logger.log('INFO', 'Starting column:', column_name)
            # 1) Set params for the actual column
            self._store_settings(params)
            # 2) By date with optional pagination (that is handled separately)
            if self._archive_page_urls_by_date:
                # a) Unique the generated archive page URLs using every day from date_from to the end of date_until
                archive_page_urls = list(set(self._gen_url_from_date(self._date_from + timedelta(days=curr_day),
                                                                     self._archive_url_format)
                                             for curr_day in range((self._date_until - self._date_from).days + 1)))
                # b) Sort the generated archive page URLs
                archive_page_urls.sort(reverse=self._go_reverse_in_archive)
            # 3) Stored in groups represented by pagination only which will be handled separately
            else:
                archive_page_urls = [self._archive_url_format]  # Only the base URL is added

            # 4) Iterate the archive URLs and process them, while generating the required page URLs on demand
            for archive_page_url in archive_page_urls:
                yield from self._gen_article_urls_including_subpages(archive_page_url)

    @staticmethod
    def _gen_url_from_date(curr_date, url_format):
        """
            Generates the URLs of a page that contains URLs of articles published on that day.
            This function allows URLs to be grouped by years or month as there is no guarantee that all fields exists.
            We also enable using open ended interval of dates. eg. from 2018-04-04 to 2018-04-05 (not included)
             or with month 2018-04-04 to 2018-05-04 (not included)
            One must place #year #month #day and #next-year #next-month #next-day labels into the url_format variable.
        """
        if '#next-day' in url_format:
            # Plus one day (open ended interval): vs.hu, hvg.hu
            next_date = curr_date + timedelta(days=1)
        elif '#next-month' in url_format:
            # Plus one month (open interval): magyarnarancs.hu
            days_in_curr_month = monthrange(curr_date.year, curr_date.month)[1]
            next_date = curr_date + timedelta(days=days_in_curr_month - curr_date.day + 1)
        else:
            # Plus one year (open interval): ???
            next_date = curr_date + timedelta(days=365 + int(isleap(curr_date.year))
                                              - curr_date.timetuple().tm_yday + 1)

        art_list_url = url_format.\
            replace('#year', '{0:04d}'.format(curr_date.year)).\
            replace('#month', '{0:02d}'.format(curr_date.month)).\
            replace('#day', '{0:02d}'.format(curr_date.day)). \
                                                              \
            replace('#next-year', '{0:04d}'.format(next_date.year)). \
            replace('#next-month', '{0:02d}'.format(next_date.month)). \
            replace('#next-day', '{0:02d}'.format(next_date.day))
        return art_list_url

    def _gen_article_urls_including_subpages(self, archive_page_url_base):
        """
            Generates article URLs from a supplied URL inlcuding the on-demand sub-pages that contains article URLs
        """
        page_num = self._min_pagenum
        first_page = True
        next_page_url = archive_page_url_base.replace('#pagenum', self._initial_page_num)
        while next_page_url is not None:
            archive_page_raw_html = self._downloader.download_url(next_page_url, self._ignore_archive_cache)
            curr_page_url = next_page_url
            if archive_page_raw_html is not None:  # Download succeeded
                self._good_urls_add(next_page_url)
                # 1) We need article URLs here to reliably determine the end of pages in some cases
                article_urls = self._extract_article_urls_from_page_fun(archive_page_raw_html)
                if len(article_urls) == 0 and (not self._infinite_scrolling or first_page):
                    self._logger.log('WARNING', next_page_url, 'Could not extract URLs from the archive!', sep='\t')
                # 2) Generate next-page URL or None if there should not be any
                next_page_url = self._find_next_page_url(archive_page_url_base, page_num, archive_page_raw_html,
                                                         article_urls)
            else:  # Download failed
                if next_page_url not in self.bad_urls and next_page_url not in self._downloader.good_urls and \
                        next_page_url not in self._downloader.url_index:  # URLs in url_index should not be a problem
                    self._problematic_urls_add(next_page_url)  # New possibly bad URL
                next_page_url = None
                article_urls = []
            page_num += 1
            self._logger.log('DEBUG', 'URLs/ARCHIVE PAGE', curr_page_url, len(article_urls), sep='\t')
            yield from article_urls
            first_page = False

    @staticmethod
    def _find_next_page_url_factory(extract_next_page_url_fun, next_url_by_pagenum, infinite_scrolling, max_pagenum,
                                    art_url_threshold, known_article_urls):

        def find_nex_page_url_spec(archive_page_url_base, page_num, raw_html, article_urls):
            """
                The next URL can be determined by various conditions (no matter how the pages are grouped):
                 1) If there is no pagination we return None
                 2) If there is a "next page" link, we find it and use that
                 3) If there is "infinite scrolling", we use pagenum from base to infinity (=No article URLs detected)
                 4) If there is only page numbering, we use pagenum from base to config-specified maximum
                 5) If there is only page numbering, we expect the archive to move during crawling (or partial crawling)
            """
            # Method #1: No pagination (default) or no page left
            next_page_url = None
            # Method #2: Use special function to follow the link to the next page
            if extract_next_page_url_fun is not None:
                next_page_url = extract_next_page_url_fun(raw_html)
            elif (next_url_by_pagenum and  # There are page numbering
                    # Method #3: No link, but infinite scrolling! (also good for inactive archive, without other clues)
                    ((infinite_scrolling and len(article_urls) > 0) or
                     # Method #4: Has predefined max_pagenum! (also good for inactive archive, with known max_pagenum)
                     (max_pagenum is not None and page_num <= max_pagenum) or
                     # Method #5: Active archive, just pages -> We allow intersecting elements
                     #  as the archive may have been moved
                     (art_url_threshold is not None and
                      (len(known_article_urls) == 0 or
                       len(article_urls.minus(known_article_urls)) > art_url_threshold)))):
                next_page_url = archive_page_url_base.replace('#pagenum', str(page_num))  # must generate URL

            return next_page_url
        return find_nex_page_url_spec


class NewsArchiveDummyCrawler:
    def __init__(self, url_index_keys, *_, **__):
        self._url_index_keys = url_index_keys

    def url_iterator(self):
        return self._url_index_keys


class NewsArticleCrawler:
    """
        1) Get the list of articles (eg. NewsArchiveCrawler)
        2) Download article pages
        3) Extract the text of articles from raw HTML
        4) Filter articles by date (depends on corpus converter)
        5) Save them in corpus format (depends on corpus converter)
        6) Follow the links on page (depends on corpus converter)
    """

    def __init__(self, settings, articles_existing_warc_filenames, articles_new_warc_filename,
                 archive_existing_warc_filenames, archive_new_warc_filename, articles_just_cache=False,
                 archive_just_cache=False, known_article_urls=None, debug_params=None, download_params=None):

        # Initialise the logger
        self._logger = Logger(settings['log_file_articles'])

        # Open files for writing gathered URLs if needed
        self._new_urls = set()
        self._new_good_urls_fh, self._new_urls_add = \
            add_and_write_factory(self._new_urls, settings.get('new_good_urls'))

        self.problematic_article_urls = set()
        self._new_problematic_urls_fh, self._problematic_article_urls_add = \
            add_and_write_factory(self.problematic_article_urls, settings.get('new_problematic_urls'))

        # Store values at init-time
        self._filter_by_date = settings['FILTER_ARTICLES_BY_DATE']
        if self._filter_by_date:  # Global date intervals, only if they are explicitly set!
            self._date_from = settings['date_from']
            self._date_until = settings['date_from']

        # Get the initialised corpus converter (can be dummy) and set the apropriate logger
        self._converter = settings['CORPUS_CONVERTER']
        self._converter.logger = self._logger

        # Create new archive while downloading, or simulate download and read the archive
        self._downloader = WarcCachingDownloader(articles_existing_warc_filenames, articles_new_warc_filename,
                                                 self._logger, articles_just_cache, download_params)

        if known_article_urls is None:  # If None is supplied copy the ones from the article archive
            known_article_urls = self._downloader.url_index  # All URLs in the archive are known good!

        if archive_just_cache and articles_just_cache:
            # Full offline mode for processing articles only withouht the archive
            self._archive_downloader = NewsArchiveDummyCrawler(self._downloader.url_index)
        else:  # known_bad_urls are common between the NewsArchiveCrawler and the NewsArticleCrawler
            # For downloading the articles from a (possibly read-only) archive
            self._archive_downloader = NewsArchiveCrawler(settings, archive_existing_warc_filenames,
                                                          archive_new_warc_filename, archive_just_cache,
                                                          known_article_urls, debug_params, download_params)

    def __del__(self):
        if hasattr(self, '_archive_downloader'):  # Make sure that the previous files are closed...
            del self._archive_downloader

        if hasattr(self, '_new_good_urls_fh') and hasattr(self, 'close'):
            self._new_good_urls_fh.close()

        if hasattr(self, '_new_problematic_urls_fh') and hasattr(self, 'close'):
            self._new_problematic_urls_fh.close()

    def _is_problematic_url(self, url):
        # Explicitly marked as bad URL (either Article or Archive) OR
        # Download failed in this session and requries manual check (either Article or Archive)
        return url in self._downloader.bad_urls or url in self._archive_downloader.bad_urls or \
               url in self.problematic_article_urls or url in self._archive_downloader.problematic_urls

    def _is_processed_good_url(self, url):
        # New good URL newly downloaded (either Article or Archive)
        # We do not count old good URLs (url_index) have taken from the cache WARC (either Article or Archive)
        #  as they are needed to be copied to the target WARC!
        return url in self._downloader.good_urls or url in self._archive_downloader.good_urls

    def download_and_extract_all_articles(self):
        self.process_urls(self._archive_downloader.url_iterator())

    def process_urls(self, it):
        urls = set()
        for url in it:
            urls.add(url)
            while len(urls) > 0:
                # This loop runs only one iteration if no URLs are extracted in step (6) else it consumes them first
                url = urls.pop()
                # 1) Check if the URL is
                # 1a) Explicitly marked as bad URL (either Article or Archive) -> Skip it, only INFO log!
                if url in self._downloader.bad_urls or url in self._archive_downloader.bad_urls:
                    self._logger.log('DEBUG', url, 'Skipping URLs explicitly marked as bad!', sep='\t')
                    continue
                # 1b) Download succeded in this session either Article or Archive (duplicate)
                # 1c) Download failed in this session and requries manual check either Article or Archive (duplicate)
                elif self._is_processed_good_url(url) or \
                        url in self.problematic_article_urls or url in self._archive_downloader.problematic_urls:
                    self._logger.log('WARNING', url, 'Not processing URL, because it is an URL already'
                                                     ' encountered in this session (including the caches)'
                                                     ' or it is known to point to the portal\'s archive!', sep='\t')
                    continue

                # 2) "Download" article
                article_raw_html = self._downloader.download_url(url)
                if article_raw_html is None:  # Download failed, must be investigated!
                    self._logger.log('ERROR', url, 'Article was not processed because download failed!', sep='\t')
                    self._problematic_article_urls_add(url)  # New problematic URL for manual checking
                    continue
                self._new_urls_add(url)  # New article URLs

                # 3) Identify the site scheme of the article to be able to look up the appropriate extracting method
                scheme = self._converter.identify_site_scheme(url, article_raw_html)

                # 4) Filter: time filtering when archive page URLs are not generated by date if needed
                if self._filter_by_date:
                    # a) Retrieve the date
                    article_date = self._converter.extract_article_date(url, article_raw_html, scheme)
                    if article_date is None:
                        self._logger.log('ERROR', url, 'DATE COULD NOT BE PARSED!', sep='\t')
                        continue
                    # b) Check date interval
                    elif not self._date_from <= article_date <= self._date_until:
                        self._logger.log('WARNING', url, 'Date ({0}) is not in the specified interval: {1}-{2}'
                                                         ' didn\'t use it in the corpus'.
                                         format(article_date, self._date_from, self._date_until), sep='\t')
                        continue

                # 5) Extract text to corpus
                self._converter.article_to_corpus(url, article_raw_html, scheme)

                # 6) Extract links to other articles and check for already extracted urls (also in the archive)?
                urls_to_follow = self._converter.follow_links_on_page(url, article_raw_html, scheme)
                # Only add those which has not been already handled to avoid loops!
                urls |= {url for url in urls_to_follow
                         if not self._is_processed_good_url(url) and not self._is_problematic_url(url)}



#!/usr/bin/env python3
# -*- coding: utf-8, vim: expandtab:ts=4 -*-

import setuptools
from webarticlecurator import __version__

with open('README.md') as fh:
    long_description = fh.read()

setuptools.setup(
    name='webarticlecurator',
    version=__version__,
    author='dlazesz',  # Will warn about missing e-mail
    description='A crawler program to download content from portals (news, forums, blogs) and convert it'
                ' to the desired output format according to the configuration.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/ELTE-DH/WebArticleCurator',
    # license='GNU Lesser General Public License v3 (LGPLv3)',  # Never really used in favour of classifiers
    # platforms='any',  # Never really used in favour of classifiers
    packages=setuptools.find_packages(exclude=['tests']),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Operating System :: POSIX :: Linux',
    ],
    python_requires='>=3.6',
    install_requires=['pyyaml', 'chardet', 'requests', 'urllib3', 'warcio', 'ratelimit', 'beautifulsoup4', 'yamale',
                      'newspaper3k'],  # Newspaper3k is optional!
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'webarticlecurator=webarticlecurator.__main__:main',
        ]
    },
)



#!/usr/bin/env python3
# -*- coding: utf-8, vim: expandtab:ts=4 -*-

# Good for downloading archive index and also the actual articles in two separate row

import os
import sys
from io import BytesIO
from collections import Counter

from warcio.warcwriter import WARCWriter
from warcio.archiveiterator import ArchiveIterator
from warcio.statusandheaders import StatusAndHeaders
from warcio.exceptions import ArchiveLoadFailed

from requests import Session
from urllib.parse import urlparse, quote, urlunparse
from requests.exceptions import RequestException
from urllib3.exceptions import ProtocolError, InsecureRequestWarning, LocationParseError
from urllib3 import disable_warnings
from chardet import detect
from ratelimit import limits, sleep_and_retry

respv_str = {10: '1.0', 11: '1.1'}


class WarcCachingDownloader:
    """
        This class optionally applies the supplied existing warc archive to retrive the downloaded pages from cache
         in case of hit and writes the retrived page to the newly created archive file.
         When the requested page is not present in the archive, it downloads the page and saves it into the new archive.

        It basically wraps WarcReader and WarcDownloader classes which do the hard work.

        All parameters are wired out to the CLI and are documented there.
    """
    def __init__(self, existing_warc_filenames, new_warc_filename, _logger, just_cache=False, download_params=None):
        self._logger = _logger
        if download_params is not None:
            strict_mode = download_params.pop('strict_mode', False)
            check_digest = download_params.pop('check_digest', False)
        else:
            strict_mode = False
            check_digest = False
            download_params = {}

        self.url_index = set()
        info_record_data = None
        if existing_warc_filenames is not None:  # Setup the supplied existing warc archive file as cache
            if isinstance(existing_warc_filenames, str):  # Transform it to list
                existing_warc_filenames = [existing_warc_filenames]
            self._cached_downloads = []
            for ex_warc_filename in existing_warc_filenames:
                cached_downloads = WarcReader(ex_warc_filename, _logger, strict_mode, check_digest)
                self._cached_downloads.append(cached_downloads)
                self.url_index |= cached_downloads.url_index
                info_record_data = cached_downloads.info_record_data

        if just_cache:
            self._new_downloads = WarcDummyDownloader()
        else:
            self._new_downloads = WarcDownloader(new_warc_filename, _logger, info_record_data, **download_params)

    def download_url(self, url, ignore_cache=False):
        # 1) Check if the URL is explicitly marked as bad...
        if url in self._new_downloads.bad_urls:
            self._logger.log('WARNING', url, 'Skipping URL explicitly marked as bad!', sep='\t')
            return None
        # 2) If the URL is present in the newly created WARC file warn and skip the URL!
        elif url in self._new_downloads.good_urls:
            # 3) throw error and return None!
            self._logger.log('ERROR', 'Not processing URL, because it is already present in the WARC archive:', url)
            return None
        # 3) Check if the URL presents in the cached_content...
        elif url in self.url_index:
            # 3a) ...copy it! (from the last source WARC where the URL is found in)
            cache, reqv, resp = self.get_records(url)
            self._new_downloads.write_record(reqv, url)
            self._new_downloads.write_record(resp, url)
            # 3b) Get content even if the URL is a duplicate, because ignore_cache knows better what to do with it
            cached_content = cache.download_url(url)
        else:
            cached_content = None

        # 4) If we have the URL cached...
        if cached_content is not None:
            if not ignore_cache:
                # 4a) and we do not expliticly ignore the cache, return the cached content!
                return cached_content
            else:
                # 4b) Log that we ignored the cached_content and do noting!
                self._logger.log('INFO', 'Ignoring cached_content for URL:', url)

        # 5) Really download the URL! (url not in cached_content or cached_content is ignored)
        return self._new_downloads.download_url(url)  # Still check if the URL is already downloaded!

    def get_records(self, url):
        for cache in reversed(self._cached_downloads):
            if url in cache.url_index:
                reqv, resp = cache.get_record(url)
                break
        else:
            raise ValueError('INTERNAL ERROR: {0} not found in any supplied source WARC file,'
                             ' but is in the URL index!'.format(url))
        return cache, reqv, resp

    @property
    def bad_urls(self):  # Ready-only property for shortcut
        return self._new_downloads.bad_urls

    @property
    def good_urls(self):  # Ready-only property for shortcut
        return self._new_downloads.good_urls


class WarcDummyDownloader:
    """
        I.e. When want to use only the cache...
    """
    def __init__(self, *_, **__):
        self.bad_urls = set()
        self.good_urls = set()

    @staticmethod
    def download_url(_):
        return None

    @staticmethod
    def write_record(*_):
        return None


class WarcDownloader:
    """
        Download URL with HTTP GET, save to a WARC file and return the decoded text
    """
    def __init__(self, expected_filename, _logger, warcinfo_record_data=None, program_name='WebArticleCurator',
                 user_agent=None, overwrite_warc=True, err_threshold=10, known_bad_urls=None,
                 max_no_of_calls_in_period=2, limit_period=1, proxy_url=None, allow_cookies=False, verify_request=True,
                 stay_offline=False):
        # Store variables
        self._logger = _logger
        self._req_headers = {'Accept-Encoding': 'identity', 'User-agent': user_agent}
        self._error_count = 0
        self._error_threshold = err_threshold  # Set the error threshold which cause aborting to prevent deinal

        # Setup download function
        if not stay_offline:
            self.download_url = self._download_url
        else:
            self.download_url = self._dummy_download_url

        if known_bad_urls is not None:  # Setup the list of cached bad URLs to prevent trying to download them again
            with open(known_bad_urls, encoding='UTF-8') as fh:
                self.bad_urls = {line.strip() for line in fh}
        else:
            self.bad_urls = set()

        self.good_urls = set()

        # Setup target file handle
        filename = self._set_target_filename(expected_filename, overwrite_warc)
        self._logger.log('INFO', 'Creating archivefile:', filename)
        self._output_file = open(filename, 'wb')

        self._session = Session()  # Setup session for speeding up downloads
        if proxy_url is not None:  # Set socks proxy if provided
            self._session.proxies['http'] = proxy_url
            self._session.proxies['https'] = proxy_url

        self._allow_cookies = allow_cookies
        self._verify_request = verify_request
        if not self._verify_request:
            disable_warnings(InsecureRequestWarning)

        # Setup rate limiting to prevent hammering the server
        self._requests_get = sleep_and_retry(limits(calls=max_no_of_calls_in_period,
                                                    period=limit_period)(self._http_get_w_cookie_handling))

        self._writer = WARCWriter(self._output_file, gzip=True, warc_version='WARC/1.1')
        if warcinfo_record_data is None:  # Or use the parsed else custom headers will not be copied
            # INFO RECORD
            # Some custom information about the warc writer program and its settings
            warcinfo_record_data = {'software': program_name, 'arguments': ' '.join(sys.argv[1:]),
                                    'format': 'WARC File Format 1.1',
                                    'conformsTo': 'http://bibnum.bnf.fr/WARC/WARC_ISO_28500_version1-1_latestdraft.pdf'}
        info_record = self._writer.create_warcinfo_record(filename, warcinfo_record_data)
        self._writer.write_record(info_record)

    @staticmethod
    def _set_target_filename(filename, overwrite_warc):
        if not overwrite_warc:  # Find out next nonexisting warc filename
            num = 0
            while os.path.exists(filename):
                filename2, ext = os.path.splitext(filename)  # Should be filename.warc.gz
                if ext == '.gz' and filename2.endswith('.warc'):
                    filename2, ext2 = os.path.splitext(filename2)  # Should be filename.warc
                    ext = ext2 + ext  # Should be .warc.gz

                filename = '{0}-{1:05d}{2}'.format(filename2, num, ext)
                num += 1
        return filename

    def __del__(self):
        if hasattr(self, '_output_file'):  # If the program opened a file, then it should gracefully close it on exit!
            self._output_file.close()

    def _http_get_w_cookie_handling(self, *args, **kwargs):
        """
            Extend requests.get with optional cookie purging
        """
        if not self._allow_cookies:
            self._session.cookies.clear()
        return self._session.get(*args, **kwargs)

    def _handle_request_exception(self, url, msg):
        self._logger.log('WARNING', url, msg, sep='\t')

        self._error_count += 1
        if self._error_count >= self._error_threshold:
            raise NameError('Too many error happened! Threshold exceeded! See log for details!')

    @staticmethod
    def _get_peer_name(resp):
        # Must get peer_name before the content is read
        # It has no official API for that:
        # https://github.com/kennethreitz/requests/issues/2158
        # https://github.com/urllib3/urllib3/issues/1071
        # So workaround to be compatible with windows:
        # https://stackoverflow.com/questions/22492484/how-do-i-get-the-ip-address-from-a-http-request-using-the-\
        # requests-library/22513161#22513161
        try:
            peer_name = resp.raw._connection.sock.getpeername()[0]  # Must get peer_name before the content is read
        except AttributeError:  # On Windows there is no getpeername() Attribute of the class...
            try:
                peer_name = resp.raw._connection.sock.socket.getpeername()[0]
            except AttributeError:
                peer_name = 'None'  # Socket closed and could not derermine peername...
        return peer_name

    def _dummy_download_url(self, _):
        raise NotImplementedError

    def _download_url(self, url):
        if url in self.bad_urls:
            self._logger.log('DEBUG', 'Not downloading known bad URL:', url)
            return None

        if url in self.good_urls:  # This should not happen!
            self._logger.log('ERROR', 'Not downloading URL, because it is already downloaded in this session:', url)
            return None

        scheme, netloc, path, params, query, fragment = urlparse(url)
        # For safety urlencode the generated URL... (The URL might by modified in this step.)
        path = quote(path, safe='/%')
        url_reparsed = urlunparse((scheme, netloc, path, params, query, fragment))

        try:  # The actual request (on the reparsed URL, everything else is made on the original URL)
            resp = self._requests_get(url_reparsed, headers=self._req_headers, stream=True, verify=self._verify_request)
        # UnicodeError is originated from idna codec error, LocationParseError is originated from URLlib3 error
        except (UnicodeError, RequestException, LocationParseError) as err:
            self._handle_request_exception(url, 'RequestException happened during downloading: {0} \n\n'
                                                ' The program ignores it and jumps to the next one.'.format(err))
            return None

        if resp.status_code != 200:  # Not HTTP 200 OK
            self._handle_request_exception(url, 'Downloading failed with status code: {0} {1}'.format(resp.status_code,
                                                                                                      resp.reason))
            return None

        # REQUEST (build headers for warc)
        reqv_headers = resp.request.headers
        reqv_headers['Host'] = netloc

        proto = 'HTTP/{0}'.format(respv_str[resp.raw.version])  # Friendly protocol name
        reqv_http_headers = StatusAndHeaders('GET {0} {1}'.format(urlunparse(('', '', path, params, query, fragment)),
                                                                  proto), reqv_headers.items(), is_http_request=True)
        reqv_record = self._writer.create_warc_record(url, 'request', http_headers=reqv_http_headers)

        # RESPONSE
        # resp_status need to be stripped else warcio strips the spaces and digest verification will fail!
        resp_status = '{0} {1}'.format(resp.status_code, resp.reason).strip()
        resp_headers_list = resp.raw.headers.items()  # get raw headers from urllib3
        # Must get peer_name before the content is read
        peer_name = self._get_peer_name(resp)

        try:
            data = resp.raw.read()  # To be able to return decoded and also write warc
        except ProtocolError as err:
            self._handle_request_exception(url, 'RequestException happened during downloading: {0} \n\n'
                                                ' The program ignores it and jumps to the next one.'.format(err))
            return None

        if len(data) == 0:
            err = 'Response data has zero length!'
            self._handle_request_exception(url, 'RequestException happened during downloading: {0} \n\n'
                                                ' The program ignores it and jumps to the next one.'.format(err))
            return None

        # warcio hack as \r\n is the record separator and trailing ones will be split and digest will eventually fail!
        if data.endswith(b'\r\n'):  # TODO: Warcio bugreport!
            data = data.rstrip()

        enc = resp.encoding  # Get or detect encoding to decode the bytes of the text to str
        if enc is None:
            enc = detect(data)['encoding']
        try:
            text = data.decode(enc)  # Normal decode process
        except UnicodeDecodeError:
            self._logger.log('WARNING', 'DECODE ERROR RETRYING IN \'IGNORE\' MODE:', url, enc, sep='\t')
            text = data.decode(enc, 'ignore')
        data_stream = BytesIO(data)  # Need the original byte stream to write the payload to the warc file

        resp_http_headers = StatusAndHeaders(resp_status, resp_headers_list, protocol=proto)
        # Add extra headers like encoding because it is not stored any other way...
        resp_record = self._writer.create_warc_record(url, 'response', payload=data_stream,
                                                      http_headers=resp_http_headers,
                                                      warc_headers_dict={'WARC-IP-Address': peer_name,
                                                                         'WARC-X-Detected-Encoding': enc})
        # Everything is OK, write the two WARC records
        self.write_record(reqv_record, url)
        self.write_record(resp_record, url)

        return text

    def write_record(self, record, url):
        self.good_urls.add(url)
        self._writer.write_record(record)


class WarcReader:
    def __init__(self, filename, _logger, strict_mode=False, check_digest=False):
        self.filename = filename
        self._stream = open(filename, 'rb')
        self._internal_url_index = {}
        self._logger = _logger
        self.info_record_data = None
        self._strict_mode = strict_mode
        if check_digest:
            check_digest = 'raise'
        self._check_digest = check_digest
        try:
            self._create_index()
        except KeyError as e:
            if self._strict_mode:
                raise e
            self._logger.log('ERROR', 'Ignoring exception:', e)

    def __del__(self):
        if hasattr(self, '_stream'):  # If the program opened a file, then it should gracefully close it on exit!
            self._stream.close()

    @property
    def url_index(self):  # Ready-only property for shortcut
        return self._internal_url_index.keys()

    def _create_index(self):
        self._logger.log('INFO', 'Creating index for {0}...'.format(self.filename))
        archive_it = ArchiveIterator(self._stream, check_digests=self._check_digest)
        info_rec = next(archive_it)
        # First record should be an info record, then it should be followed by the request-response pairs
        assert info_rec.rec_type == 'warcinfo'
        try:
            # Read out custom headers for later use
            custom_headers_raw = info_rec.content_stream().read()  # Parse custom headers
            if len(custom_headers_raw) == 0:
                raise ValueError('WARCINFO record payload length is 0!')
            # Read and parse the warcinfo record for writing it back unchanged into a warc file
            # else due to warcio problems it will not be copied properly!
            # See: https://github.com/webrecorder/warcio/issues/90
            # and https://github.com/webrecorder/warcio/issues/91
            self.info_record_data = dict(r.split(': ', maxsplit=1) for r in custom_headers_raw.decode('UTF-8')
                                         .strip().split('\r\n') if len(r) > 0)
        except ValueError as e:
            if self._strict_mode:
                raise e
            self._logger.log('WARNING', 'WARCINFO record in', self._stream.name,
                             'is corrupt! Continuing with a fresh one!')
            self.info_record_data = None

        archive_load_failed = False
        count = 0
        double_urls = Counter()
        reqv_data = (None, (None, None))  # To be able to handle the request-response pairs together
        for i, record in enumerate(archive_it):
            if record.rec_type == 'request':
                assert i % 2 == 0
                try:
                    reqv_data = (record.rec_headers.get_header('WARC-Target-URI'),
                                 (archive_it.get_record_offset(), archive_it.get_record_length()))
                except ArchiveLoadFailed as e:
                    self._logger.log('ERROR', 'REQUEST:', e.msg, 'for', reqv_data[0])
                    archive_load_failed = True
            if record.rec_type == 'response':
                assert i % 2 == 1
                resp_url = record.rec_headers.get_header('WARC-Target-URI')
                assert resp_url == reqv_data[0]
                double_urls[resp_url] += 1
                try:
                    self._internal_url_index[resp_url] = (reqv_data[1],  # Request-response pair
                                                          (archive_it.get_record_offset(),
                                                           archive_it.get_record_length()))
                except ArchiveLoadFailed as e:
                    self._logger.log('ERROR', 'RESPONSE:', e.msg, 'for', resp_url)
                    archive_load_failed = True
                count += 1
        if count != len(self._internal_url_index):
            double_urls_str = '\n'.join('{0}\t{1}'.format(url, freq) for url, freq in double_urls.most_common()
                                        if freq > 1)
            raise KeyError('The following double URLs detected in the WARC file:{0}'.format(double_urls_str))
        if count == 0:
            raise IndexError('No index created or no response records in the WARC file!')
        if archive_load_failed and self._strict_mode:
            raise ArchiveLoadFailed('Archive loading failed! See logs for details!')
        self._stream.seek(0)
        self._logger.log('INFO', 'Index succesuflly created.')

    def get_record(self, url):
        reqv_resp_pair = self._internal_url_index.get(url)
        if reqv_resp_pair is not None:
            self._stream.seek(reqv_resp_pair[0][0])
            reqv = next(iter(ArchiveIterator(self._stream, check_digests=self._check_digest)))
            self._stream.seek(reqv_resp_pair[1][0])
            resp = next(iter(ArchiveIterator(self._stream, check_digests=self._check_digest)))
            return reqv, resp
        else:
            raise KeyError('The request or response is missing from the archive for URL: {0}'.format(url))

    def download_url(self, url):
        text = None
        reqv_resp_pair = self._internal_url_index.get(url)
        if reqv_resp_pair is not None:
            offset = reqv_resp_pair[1][0]  # Only need the offset of the response part
            self._stream.seek(offset)  # Can not be cached as we also want to write it out to the new archive!
            record = next(iter(ArchiveIterator(self._stream, check_digests=self._check_digest)))
            data = record.content_stream().read()
            assert len(data) > 0
            enc = record.rec_headers.get_header('WARC-X-Detected-Encoding', 'UTF-8')
            text = data.decode(enc, 'ignore')
        else:
            self._logger.log('CRITICAL', url, 'URL not found in WARC!', sep='\t')

        return text


def sample_warc_by_urls(source_warcfiles, new_urls, sampler_logger, target_warcfile=None, out_dir=None, offline=True,
                        just_cache=False):
    """ Create new warc file for the supplied list of URLs from an existing warc file """
    is_out_dir_mode = out_dir is not None
    if is_out_dir_mode:
        os.makedirs(out_dir, exist_ok=True)
        if len(os.listdir(out_dir)) != 0:
            print('Supplied output directory ({0}) is not empty!'.format(out_dir), file=sys.stderr)
            exit(1)

    w = WarcCachingDownloader(source_warcfiles, target_warcfile, sampler_logger, just_cache=just_cache,
                              download_params={'stay_offline': offline})
    for url in new_urls:
        url = url.strip()
        sampler_logger.log('INFO', 'Adding url', url)
        cont = w.download_url(url)
        if is_out_dir_mode and cont is not None:
            safe_url = ''.join(char if char.isalnum() else '_' for char in url).rstrip('_')[:200]
            i, fname = 0, os.path.join(out_dir, '{0}{1}.html'.format(safe_url, 0))
            while os.path.exists(fname):
                i += 1
                fname = os.path.join(out_dir, '{0}{1}.html'.format(safe_url, i))
            sampler_logger.log('INFO', 'Creating file', fname)
            with open(fname, 'w', encoding='UTF-8') as fh:
                fh.write(cont)


def validate_warc_file(source_warcfiles, validator_logger):
    reader = WarcCachingDownloader(source_warcfiles, None, validator_logger, True,
                                   download_params={'stay_offline': True, 'strict_mode': True, 'check_digest': True})
    validator_logger.log('INFO', 'OK!', len(reader.url_index), 'records read!')
    return reader.url_index


def online_test(url='https://index.hu/belfold/2018/08/27/fidesz_media_helyreigazitas/', filename='example.warc.gz',
                test_logger=None):
    w = WarcCachingDownloader(None, filename, test_logger)
    t = w.download_url(url)
    test_logger.log('INFO', t)



#!/usr/bin/env python3
# -*- coding: utf-8, vim: expandtab:ts=4 -*-

from webarticlecurator.logger import Logger
from webarticlecurator.utils import wrap_input_constants, DummyConverter
from webarticlecurator.enhanced_downloader import WarcCachingDownloader, sample_warc_by_urls, validate_warc_file, \
    online_test
from webarticlecurator.news_crawler import NewsArchiveCrawler, NewsArticleCrawler
from webarticlecurator.version import __version__

__all__ = ['NewsArchiveCrawler', 'NewsArticleCrawler', 'DummyConverter', 'WarcCachingDownloader', 'sample_warc_by_urls',
           'validate_warc_file', 'online_test', 'Logger', 'wrap_input_constants', __version__]
